/*
  based on sketch: https://github.com/aknik/ESP32/blob/master/DFC77/DFC77_esp32_Solo.ino

  Some functions are inspired by work of G6EJD ( https://www.youtube.com/channel/UCgtlqH_lkMdIa4jZLItcsTg )

  Refactor by DeltaZero, converts to syncronous, added "cron" that you can bypass, see line 29
*/

// modified by PeterN to WiFi-Manager and semi-continuous mode with NeoPixel feedback on M5_ATOM_lite
// VERSION 2023/03/27/a

//======================================================================================================================================
// CRON-mode deaktivated    PeterN
#define CONTINUOUSMODE  // to bypass de cron and have the transmitter on all the time
// continuous mode  needs regular reset to sync time with NTP-Server
#define minuteToAutoReset 5  // reset after x minutes in continuous mode, because NTP-Sync happens only at startup     PeterN
// BUT: some clocks need a time >10m without dcf-signal to ensure a new sync more often
#define minuteWithoutDFC 15  // Minutes without any DCF  output signal following the minuteToAutoReset period          PeterN
//======================================================================================================================================

#include <WiFiManager.h>  //peterN
#include <WiFi.h>
#include <Ticker.h>
#include <Time.h>

// buitin NeoPixel in an M5Stack ATOM lite used as feedbacK LED     PeterN
#include <Adafruit_NeoPixel.h>
#define PIN 27
#define NUMPIXELS 1
Adafruit_NeoPixel pixels = Adafruit_NeoPixel(NUMPIXELS, PIN, NEO_GRB + NEO_KHZ800);

const char* ssid = "";      // empty as WiFiManager does the job now    //PeterN
const char* password = "";  // empty as WiFiManager does the job now    //PeterN

#define DCF_OUT_1 26  // M5Stack ATOM lite groove port GPio26 as output to the clock    PeterN
#define DCF_OUT_2 32  // M5Stack ATOM lite groove port GPio32 as ADDITIONAL output to a SECOND clock for testing     PeterN

#define ANTENNAPIN 25  // M5Stack ATOM lite  GPio25 as 77.5kHz RF-output     PeterN
#define BUTTON     39  //Builtin button of M5Stack ATOM to force WiFiManager in AP-mode at startup    PeterN
#define TZ_PIN     21  // input to enforce australian timezone if pulled to GND 

// cron (if you choose the correct values you can even run on batteries)
// If you choose really bad this minutes, everything goes wrong, so minuteToWakeUp must be greater than minuteToSleep
#define minuteToWakeUp 55  // Every hoursToWakeUp at this minute the ESP32 wakes up get time and starts to transmit
#define minuteToSleep 8    // If it is running at this minute then goes to sleep and waits until minuteToWakeUp

byte hoursToWakeUp[] = { 0, 1, 2, 3 };  // you can add more hours to adapt to your needs
                                        // When the ESP32 wakes up, check if the actual hour is in the list and
                                        // runs or goes to sleep until next minuteToWakeUp

Ticker tickerDecisec;  // TBD at 100ms

//complete array of pulses for a minute
//0 = no pulse, 1=100ms, 2=200ms
int impulseArray[60];
int impulseCount = 0;
int actualHours, actualMinutes, actualSecond, actualDay, actualMonth, actualYear, DayOfW;
long dontGoToSleep = 0;
const long onTimeAfterReset = 600000;  // Ten minutes
int timeRunningContinuous = 0;

const char* ntpServer = "1.de.pool.ntp.org";                           // enter your closer pool or pool.ntp.org
const char* TZ_INFO = "CET-1CEST-2,M3.5.0/02:00:00,M10.5.0/03:00:00";  // enter your time zone (https://remotemonitoringsystems.ca/time-zone-abbreviations.php)
const char* TZ_INFO_TEST = "EST-10EDT-11,M10.5.0/02:00:00,M3.5.0/03:00:00";  // AUSTRALIA enter your time zone (https://remotemonitoringsystems.ca/time-zone-abbreviations.php)

struct tm timeinfo;

void setup() {
  esp_sleep_disable_wakeup_source(ESP_SLEEP_WAKEUP_ALL);
  //Serial.begin(115200);
  Serial.begin(250000);

  if (esp_sleep_get_wakeup_cause() == 0) dontGoToSleep = millis();

  ledcSetup(0, 77500, 8);        // DCF77 frequency
  ledcAttachPin(ANTENNAPIN, 0);  // This Pin, or another one you choose, has to be attached to the antenna

  pinMode(DCF_OUT_1, OUTPUT);
  digitalWrite(DCF_OUT_1, LOW);  // LOW if DCF_OUT_1  is inverted like in Wemos boards
  pinMode(DCF_OUT_2, OUTPUT);
  digitalWrite(DCF_OUT_2, LOW);  // LOW if DCF_OUT_2  is inverted like in Wemos boards

  pixels.begin();                                    // Neopixel as feedback    PeterN
  pixels.setPixelColor(0, pixels.Color(10, 10, 0));  // Neopixel = YELLOW  while WiFi Manager is active      PeterN
  pixels.show();                                     // PeterN

  pinMode(BUTTON, INPUT_PULLUP);  // build-in pushbutton, active low
  pinMode(TZ_PIN, INPUT_PULLUP);  // input-pin to enforce timezone of SINDEY, active low
  
  //###PeterN #####################################################################################################################
  //WiFiManager, Local intialization. Once its business is done, there is no need to keep it around
  WiFiManager wm;
  wm.setConfigPortalTimeout(5 * 60);  //timeout for WiFi-Manager captive portal    //PeterN

  // RESET  WIFI-CREDENTIALS BY PRESSING FRONTBUTTON for > 3s AT STARTUP
  // THIS WILL ENFORCE AP-MODE WITH A CAPTIVE PORTAL FOR NEW WIFI-CREDENTIALS
  if ((digitalRead(BUTTON) == 0)) {  // If build-in button is pressed
    delay(3000);
    if ((digitalRead(BUTTON) == 0)) {  // If build-in button is still pressed

      Serial.println("Pushbutton was pressed while start >>> WiFi-credentials deleted. WiFi AccessPoint with captive Portal will start now. ");
      pixels.setPixelColor(0, pixels.Color(0, 10, 80));  // Neopixel = strong blue  while WiFiManager gets credentials   PeterN
      pixels.show();                                     // PeterN
      delay(3000);                                       // PeterN
      wm.resetSettings();                                // PeterN
    }
  }

  //###PeterN #####################################################################################################################
  //WiFiManager, Local intialization. Once its business is done, there is no need to keep it around
  //WiFiManager wm;

  bool res;
  res = wm.autoConnect("DCF77_ESP32", "");  // AP with no password

  if (!res) {
    Serial.println("Failed to connect to stored WiFi . RESTART ");
    pixels.setPixelColor(0, pixels.Color(10, 0, 0));  // Neopixel = RED   PeterN
    pixels.show();                                    // Neopixel = RED as feedback   PeterN
    delay(5000);
    ESP.restart();
  } else {
    //if you get here you have connected to the WiFi
    Serial.println("Now connected to a WiFi ########################");
    pixels.setPixelColor(0, pixels.Color(5, 5, 5));  // Neopixel = white   PeterN
    pixels.show();                                   // Neopixel = white as feedback   PeterN
  }

  //###PeterN #####################################################################################################################
  pixels.setPixelColor(0, pixels.Color(10, 10, 10));  // WHITE   PeterN
  pixels.show();                                      // PeterN

  ////////WiFi_on();       //deaktivated by PeterN because WiFiManager already did this job
  getNTP();  // only once here, and not again until restart

  Serial.print("   wm.stopWebPortal() ");
  wm.stopWebPortal();

  Serial.print("   wm.stopConfigPortal() ");
  res = wm.stopConfigPortal();
  Serial.print(res);

  Serial.print("   wm.disconnect() ");
  res = wm.disconnect();
  Serial.println(res);

  WiFi_off();

  Serial.println();
  Serial.println("#### DCF77-transmitter now starting up.####");

  show_time();

  CodeTime();                                                      // first conversion just for cronCheck
  //++workaround for special sensitive clocks++++++++++++++++++++++++++++++++++++
  Serial.println("...Now waiting for 58th second of current minute.");  //PeterN
  do {
    // This loop just  prevents from a start with a partitial first DCF-telegram       PeterN
    pixels.setPixelColor(0, pixels.Color(60 - timeinfo.tm_sec, timeinfo.tm_sec, 0));  // optical countdown from red to green  PeterN
    pixels.show();
    if (!getLocalTime(&timeinfo)) {
      Serial.println("Error obtaining time...");
      delay(3000);
      ESP.restart();
    }
  } while (timeinfo.tm_sec < 58);
  Serial.println("OK: 58th second of current minute  reached.");  //PeterN

  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


#ifndef CONTINUOUSMODE
  if ((dontGoToSleep == 0) or ((dontGoToSleep + onTimeAfterReset) < millis())) cronCheck();  // first check before start anything
#else
  Serial.print("CONTINUOUS MODE NO CRON!!!  Next NTP-Sync will be after reboot in ");  //PeterN
  Serial.print(minuteToAutoReset);                                                     //PeterN
  Serial.println(" minutes.");                                                         //PeterN
#endif

  // sync to the start of a second
  Serial.print("Syncing to start of next second... ");
  int startSecond = timeinfo.tm_sec;
  long count = 0;
  do {
    count++;
    if (!getLocalTime(&timeinfo)) {
      Serial.println("Error obtaining time...");
      delay(3000);
      ESP.restart();
    }
  } while (startSecond == timeinfo.tm_sec);

  tickerDecisec.attach_ms(100, DcfOut);  // from now on calls DcfOut every 100ms
                                         // ATTENTION: delay() will deactiate ticker !!
  Serial.print("NTP Ok    ");
  Serial.println(count);
}

//#########################################################
void loop() {
  // There is no code inside the loop. This is a syncronous program driven by the Ticker
}
//#########################################################

//####################################################################################################
void CodeTime() {
  DayOfW = timeinfo.tm_wday;
  if (DayOfW == 0) DayOfW = 7;
  actualDay = timeinfo.tm_mday;
  actualMonth = timeinfo.tm_mon + 1;
  actualYear = timeinfo.tm_year - 100;
  actualHours = timeinfo.tm_hour;
  actualMinutes = timeinfo.tm_min + 1;  // DCF77 transmitts the next minute
  if (actualMinutes >= 60) {
    actualMinutes = 0;
    actualHours++;
  }
  actualSecond = timeinfo.tm_sec;
  if (actualSecond == 60) actualSecond = 0;

  int n, Tmp, TmpIn;
  int ParityCount = 0;

  //we put the first 20 bits of each minute at a logical zero value
  for (n = 0; n < 20; n++) impulseArray[n] = 1;

  // set DST bit
  if (timeinfo.tm_isdst == 0) {
    impulseArray[18] = 2;  // CET or DST OFF
  } else {
    impulseArray[17] = 2;  // CEST or DST ON
  }

  //bit 20 must be 1 to indicate active time
  impulseArray[20] = 2;

  //calculates the bits for the minutes
  TmpIn = Bin2Bcd(actualMinutes);
  for (n = 21; n < 28; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  if ((ParityCount & 1) == 0)
    impulseArray[28] = 1;
  else
    impulseArray[28] = 2;

  //calculates bits for the hours
  ParityCount = 0;
  TmpIn = Bin2Bcd(actualHours);
  for (n = 29; n < 35; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  if ((ParityCount & 1) == 0)
    impulseArray[35] = 1;
  else
    impulseArray[35] = 2;
  ParityCount = 0;

  //calculate the bits for the actual Day of Month
  TmpIn = Bin2Bcd(actualDay);
  for (n = 36; n < 42; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  TmpIn = Bin2Bcd(DayOfW);
  for (n = 42; n < 45; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  //calculates the bits for the actualMonth
  TmpIn = Bin2Bcd(actualMonth);
  for (n = 45; n < 50; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  //calculates the bits for actual year
  TmpIn = Bin2Bcd(actualYear);  // 2 digit year
  for (n = 50; n < 58; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  //equal date
  if ((ParityCount & 1) == 0)
    impulseArray[58] = 1;
  else
    impulseArray[58] = 2;

  //last missing pulse
  impulseArray[59] = 0;  // No pulse
}

//####################################################################################################
int Bin2Bcd(int dato) {
  int msb, lsb;
  if (dato < 10)
    return dato;
  msb = (dato / 10) << 4;
  lsb = dato % 10;
  return msb + lsb;
}
//####################################################################################################
void DcfOut() {
  switch (impulseCount++) {
    case 0:
      if (impulseArray[actualSecond] != 0) {
        //digitalWrite(DCF_OUT_1 , LOW);   //original      PeterN
        digitalWrite(DCF_OUT_1, HIGH);  //inverted      PeterN
        //digitalWrite(DCF_OUT_2, HIGH);  //inverted      PeterN
        ledcWrite(0, 0);
        pixels.setPixelColor(0, pixels.Color(0, 8, 0));  // Neopixel as feedback   PeterN
        pixels.show();                                   // Neopixel as feedback   PeterN
      }
      break;
    case 1:
      if (impulseArray[actualSecond] == 1) {
        //digitalWrite(DCF_OUT_1 , HIGH);  //original       PeterN
        digitalWrite(DCF_OUT_1, LOW);  //inverted       PeterN
        //digitalWrite(DCF_OUT_2, LOW);  //inverted       PeterN
        ledcWrite(0, 127);
        pixels.setPixelColor(0, pixels.Color(0, 0, 0));  // Neopixel as feedback   PeterN
        pixels.show();                                   // Neopixel as feedback   PeterN
      }
      break;
    case 2:
      //digitalWrite(DCF_OUT_1 , HIGH);   //original       PeterN
      digitalWrite(DCF_OUT_1, LOW);  //inverted       PeterN
      //digitalWrite(DCF_OUT_2, LOW);  //inverted       PeterN
      ledcWrite(0, 127);
      pixels.setPixelColor(0, pixels.Color(0, 0, 0));  // Neopixel as feedback   PeterN
      pixels.show();                                   // Neopixel as feedback   PeterN

      break;
    case 9:
      impulseCount = 0;

      if (actualSecond == 1 || actualSecond == 15 || actualSecond == 21 || actualSecond == 29) Serial.print("-");
      if (actualSecond == 36 || actualSecond == 42 || actualSecond == 45 || actualSecond == 50) Serial.print("-");
      if (actualSecond == 28 || actualSecond == 35 || actualSecond == 58) Serial.print("P");

      if (impulseArray[actualSecond] == 1) Serial.print("0");
      if (impulseArray[actualSecond] == 2) Serial.print("1");

      if (actualSecond == 59) {
        Serial.println();
        show_time();
#ifndef CONTINUOUSMODE
        if ((dontGoToSleep == 0) or ((dontGoToSleep + onTimeAfterReset) < millis())) cronCheck();
#else
        timeRunningContinuous++;
        Serial.println("SEMI-CONTINUOUS MODE. NO CRON!!! NTP-sync of ESP32 clock happens at startup only.");  //###############  PeterN
        Serial.print(minuteToAutoReset - timeRunningContinuous);                                              //###############  PeterN
        Serial.println(" minutes left to the next restart with NTP-Sync.");                                   //###############  PeterN
        if (timeRunningContinuous >= minuteToAutoReset) {                                                     // minuteToAutoReset minutes running, then restart all over  ###############  PeterN
          long count = minuteWithoutDFC * 60;                                                                 // now countdown  10 Minutes without DCF impulses until  reset
          do {
            digitalWrite(DCF_OUT_1, HIGH);                    // Force output for period without pulses to HIGH, as if no DCF-antenna is connected      PeterN
            ledcWrite(0, 0);                                  //77.5kHz off
            pixels.setPixelColor(0, pixels.Color(0, 5, 0));  // Neopixel as feedback   PeterN
            pixels.show();                                    // Neopixel as feedback   PeterN
            delay(500);                                       //delay deactivates the ticker !!! so no more DCF_pulses untill reset
            pixels.setPixelColor(0, pixels.Color(0, 10, 0));  // Neopixel as feedback   PeterN
            pixels.show();                                    // Neopixel as feedback   PeterN
            Serial.print(" DCF-output is deactivated for another ");
            Serial.print(count);
            Serial.println(" seconds. ");
            delay(500);
            count--;
          } while (count > 0);

          ESP.restart();  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        }
#endif
      }
      break;
  }
  if (!getLocalTime(&timeinfo)) {
    Serial.println("Error obtaining time...");
    delay(3000);
    ESP.restart();
  }
  CodeTime();
}
